@@ -20,6 +20,7 @@ from rdkit import Chem
 from tqdm import tqdm
 import time
 from datetime import datetime
+from filelock import FileLock
 
 from boltz.data import const
 from boltz.data.module.inference import BoltzInferenceDataModule
@@ -573,13 +574,15 @@ def process_input(  # noqa: C901, PLR0912, PLR0915, D103
             # Add to generate list, assigning entity id
             if (chain.mol_type == prot_id) and (chain.msa_id == 0):
                 entity_id = chain.entity_id
-                # msa_id = f"{target_id}_{entity_id}"
-                # to_generate[msa_id] = target.sequences[entity_id]
                 msa_id = chain.hash_seq  # f"{target_id}_{entity_id}"
                 chain.msa_id = msa_dir / f"{msa_id}.csv"
-                if not chain.msa_id.exists():
-                    to_generate[msa_id] = target.sequences[entity_id]
-
+                lock_path = str(chain.msa_id) + ".lock"
+                with FileLock(lock_path):
+                    # Check again inside the lock!
+                    if not chain.msa_id.exists() or os.path.getsize(chain.msa_id) == 0:
+                        to_generate[msa_id] = target.sequences[entity_id]
+                        # Generate MSA if needed (existing logic will handle this after loop)
+                    # else: file exists and is non-empty, do nothing (will be used)
             # We do not support msa generation for non-protein chains
             elif chain.msa_id == 0:
                 chain.msa_id = -1
@@ -592,17 +595,22 @@ def process_input(  # noqa: C901, PLR0912, PLR0915, D103
         if to_generate:
             msg = f"Generating MSA for {path} with {len(to_generate)} protein entities."
             click.echo(msg)
-            compute_msa(
-                data=to_generate,
-                target_id=target_id,
-                msa_dir=msa_dir,
-                msa_server_url=msa_server_url,
-                msa_pairing_strategy=msa_pairing_strategy,
-                msa_server_username=msa_server_username,
-                msa_server_password=msa_server_password,
-                api_key_header=api_key_header,
-                api_key_value=api_key_value,
-            )
+            for msa_id, seq in to_generate.items():
+                msa_path = msa_dir / f"{msa_id}.csv"
+                lock_path = str(msa_path) + ".lock"
+                with FileLock(lock_path):
+                    if not msa_path.exists() or os.path.getsize(msa_path) == 0:
+                        compute_msa(
+                            data={msa_id: seq},
+                            target_id=target_id,
+                            msa_dir=msa_dir,
+                            msa_server_url=msa_server_url,
+                            msa_pairing_strategy=msa_pairing_strategy,
+                            msa_server_username=msa_server_username,
+                            msa_server_password=msa_server_password,
+                            api_key_header=api_key_header,
+                            api_key_value=api_key_value,
+                        )
 
         # Parse MSA data
         # msas = sorted({c.msa_id for c in target.record.chains if c.msa_id != -1})
